# -*- coding: utf-8; py-indent-offset:4 -*-

import os, sys
import datetime as dt
import tabulate as tb
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

from ..core import get_cn_fund_list, get_cn_fund_daily, get_cn_fund_manager, get_cn_index_list_df, get_index_daily
from ..utils import dict_from_df, datetime_today, sort_with_options, filter_with_options, date_limit_from_options, csv_xlsx_from_options

def cli_fund_help():
    syntax_tips = '''Syntax:
    __argv0__ fund update <all | symbols | symbols.csv>
    __argv0__ fund list [<keyword>] [-sortby=...] [-desc] [-filter_column=...-...]
    __argv0__ fund manager [<keyword>] [-s | -d] [-sortby=...] [-desc] [-filter_column=...-...]
    __argv0__ fund company [<keyword>]
    __argv0__ fund eval <all | symbols | symbols.csv> [-sortby=...] [-desc] [-filter_column=...-...]
    __argv0__ fund plot <symbols | symbols.csv> [<options>]
    __argv0__ fund backtest  <all | symbols | symbols.csv> [-period=90] [-days=...] [-date=yyyymmdd-yyyymmdd]

Options:
    -sortby=<col> .................. sort by the column
    -sharpe=2.5- ................... sharpe value between <from> to <to>
    -max_drawdown=-20 .............. max_drawdown between <from> to <to>
    -volatility=-20 ................ volatility between <from> to <to>
    -out=file.csv .................. export fund list to .csv file
    -out=file.xlsx ................. export fund data to .xlsx file

    -s ............................. display symbol of the funds managed
    -d ............................. display symbol and name of the funds managed

Example:
    __argv0__ fund list 多因子
    __argv0__ fund list 广发 -out=output/myfunds.csv
    __argv0__ fund company 华安
    __argv0__ fund update data/myfunds.csv
    __argv0__ fund eval all -days=365 -sortby=sharpe -desc -limit=20 -out=output/top20_funds.xlsx
    __argv0__ fund plot 002943 005669 000209 -days=365
    __argv0__ fund plot data/funds.csv -days=365
    __argv0__ fund plot data/funds.csv -days=1095 -mix
'''.replace('__argv0__',os.path.basename(sys.argv[0]))

    print( syntax_tips )

def get_fund_symbol_name():
    df = get_cn_fund_list(check_date= datetime_today())
    return dict_from_df(df, 'symbol', 'name')

def get_fund_name_symbol():
    df = get_cn_fund_list(check_date= datetime_today())
    return dict_from_df(df, 'name', 'symbol')

def get_fund_manager_mapping():
    df = get_cn_fund_manager(check_date= datetime_today())
    fund_manager = {}
    for i, row in df.iterrows():
        name = row['name']
        fund = row['fund']
        if fund in fund_manager:
            fund_manager[ fund ].append( name )
            pass
        else:
            fund_manager[ fund ] = [ name ]
    return fund_manager

def get_manager_fundname_mapping():
    df = get_cn_fund_manager(check_date= datetime_today())
    manager_fund = {}
    for i, row in df.iterrows():
        name = row['company'] + ' ' + row['name']
        fund = row['fund']
        if name in manager_fund:
            manager_fund[ name ].append( fund )
        else:
            manager_fund[ name ] = [ fund ]
    return manager_fund

def get_manager_fundsymbol_mapping():
    fund_symbol = dict_from_df(get_cn_fund_list(check_date= datetime_today()), 'name', 'symbol')
    df = get_cn_fund_manager(check_date= datetime_today())
    manager_fund = {}
    for i, row in df.iterrows():
        name = row['company'] + ' ' + row['name']
        fund = row['fund']
        symbol = fund_symbol[fund] if (fund in fund_symbol) else ''
        if symbol == '':
            continue
        if name in manager_fund:
            manager_fund[ name ].append( symbol )
        else:
            manager_fund[ name ] = [ symbol ]
    return manager_fund

def get_manager_fund_mapping():
    fund_symbol = dict_from_df(get_cn_fund_list(check_date= datetime_today()), 'name', 'symbol')
    df = get_cn_fund_manager(check_date= datetime_today())
    manager_fund = {}
    for i, row in df.iterrows():
        name = row['company'] + ' ' + row['name']
        fund = row['fund']
        symbol = fund_symbol[fund] if (fund in fund_symbol) else ''
        if symbol == '':
            continue
        if name in manager_fund:
            manager_fund[ name ].append( symbol + ' - ' + fund )
        else:
            manager_fund[ name ] = [ symbol + ' - ' + fund ]
    return manager_fund

# hiquant fund list
# hiquant fund list 多因子
def cli_fund_list(params, options):
    df = get_cn_fund_list(check_date= datetime_today())
    selected = total = df.shape[0]
    if len(params) > 0:
        df = df[ df['name'].str.contains(params[0], na=False) ]
        selected = df.shape[0]

    df = filter_with_options(df, options)
    df = sort_with_options(df, options, by_default='symbol')

    limit = 0
    for option in options:
        if option.startswith('-limit='):
            limit = int(option.replace('-limit=',''))
    if limit > 0:
        df = df.head(limit)

    print( tb.tabulate(df, headers='keys') )
    print( selected, 'of', total, 'funds selected.')

    out_csv_file, out_xls_file = csv_xlsx_from_options(options)

    if out_csv_file:
        df = df[['symbol', 'name']]
        df.to_csv(out_csv_file, index= False)
        print('Exported to:', out_csv_file)
        print(df)

def get_cn_fund_company(keyword = None):
    df = get_cn_fund_manager(check_date= datetime_today())
    if keyword is not None:
        df = df[ df['company'].str.contains(keyword, na=False) ]

    mapping = {}
    for i, row in df.iterrows():
        fund = row['fund']
        manager = row['name']
        company = row['company']
        if company in mapping:
            funds = mapping[ company ]['funds']
            managers = mapping[ company ]['managers']
            if fund not in funds:
                funds.append( fund )
            if manager not in managers:
                managers.append( manager )
        else:
            mapping[ company ] = {
                'funds': [ fund ],
                'managers': [ manager ],
            }
    table = []
    for company, v in mapping.items():
        funds = v['funds']
        managers = v['managers']
        table.append( [company, len(managers), len(funds)] )
    return pd.DataFrame(table, columns=['company','managers','funds']).sort_values(by= 'funds', ascending= False).reset_index(drop= True)

def cli_fund_company(params, options):
    keyword = params[0] if len(params) > 0 else None
    df = get_cn_fund_company(keyword)

    selected = total = df.shape[0]
    limit = 0
    for option in options:
        if option.startswith('-limit='):
            limit = int(option.replace('-limit=',''))
    if limit > 0:
        df = df.head(limit)

    print( tb.tabulate(df, headers='keys') )
    print( df.shape[0], 'fund companies.')

def cli_fund_manager(params, options):
    df = get_cn_fund_manager(check_date= datetime_today())
    selected = total = df.shape[0]

    if len(params) > 0:
        keyword = params[0]
        if keyword.endswith('.csv'):
            df1 = df.copy()
            df1['manager'] = df['company'] + ' ' + df['name']
            df_manager = pd.read_csv(keyword, dtype= str)
            df_manager['manager'] = df_manager['company'] + ' ' + df_manager['name']
            df = df1[ df1['manager'].isin(df_manager['manager'].tolist()) ].drop(columns=['manager'])
        else:
            df1 = df.copy()
            df1['keywords'] = df1['company'] + df1['name'] + ' ' + df1['fund']
            df = df1[ df1['keywords'].str.contains(keyword, na=False) ].drop(columns=['keywords'])

    yeartop = 0
    limit = 0
    for k in options:
        if k.startswith('-limit='):
            limit = int(k.replace('-limit=',''))
        if k.startswith('-yeartop='):
            yeartop = int(k.replace('-yeartop=', ''))
        if k.startswith('-fund='):
            fund = k.replace('-fund=','')
            df = df[ df['fund'].str.contains(fund, na=False) ]

    df_company = get_cn_fund_company()
    company_managers = dict_from_df(df_company, 'company', 'managers')
    company_funds = dict_from_df(df_company, 'company', 'funds')

    group = '-d' not in options
    if group and (df.shape[0] > 0):
        df_tmp = df.drop(columns=['fund'])
        table = []
        name = ''
        for i, row in df_tmp.iterrows():
            company = row['company']
            manager = company + row['name']
            if name == manager:
                continue
            else:
                name = manager
                data = list(row.values)

                managers = company_managers[company] if (company in company_managers) else 0
                funds = company_funds[company] if (company in company_funds) else 0
                data.insert(2, managers)
                data.insert(3, funds)

                table.append( data )

        cols = list(row.keys())
        cols.insert(2, 'managers')
        cols.insert(3, 'funds')
        df = pd.DataFrame(table, columns=cols)

    df['annual'] = round((np.power((df['best_return'] * 0.01 + 1), 1.0/(df['days']/365.0)) - 1.0) * 100.0, 1)
    df['annual'] = df[['best_return', 'annual']].min(axis= 1)

    if yeartop > 0:
        df1 = df[ df['days'] >= 3650 ].sort_values(by='best_return', ascending=False).head(yeartop)
        for i in range(9,-1,-1):
            df2 = df[ (df['days'] >= (i*365)) & (df['days'] < ((i+1))*365) ].sort_values(by='best_return', ascending=False).head(yeartop)
            df1 = pd.concat([df1, df2], ignore_index=True)
        df1.insert(6, 'years', round(df1['days'] / 365.0, 1))
        df = df1.drop(columns= ['days'])

    df = filter_with_options(df, options)
    for k in options:
        if k.startswith('-sortby='):
            df = sort_with_options(df, options, by_default='best_return')
            break
    if limit > 0:
        df = df.head(limit)

    if 'fund' in df.columns:
        fund_name_symbol = get_fund_name_symbol()
        df['fund'] = [(fund_name_symbol[fund] if fund in fund_name_symbol else '') for fund in df['fund'].tolist()] + (' - ' + df['fund'])
    elif ('-s' in options) and ('name'in df.columns):
        manager_fundsymbol = get_manager_fundsymbol_mapping()
        managers = (df['company'] + ' ' + df['name']).tolist()
        df['symbol'] = [(','.join(manager_fundsymbol[manager]) if manager in manager_fundsymbol else '') for manager in managers]
    elif ('-sd' in options) and ('name'in df.columns):
        manager_fund = get_manager_fund_mapping()
        managers = (df['company'] + ' ' + df['name']).tolist()
        df['funds'] = [('\n'.join(manager_fund[manager]) if manager in manager_fund else '') for manager in managers]

    selected = df.shape[0]
    print( tb.tabulate(df, headers='keys') )
    print( selected, 'of', total, 'selected.')

    out_csv_file, out_xls_file = csv_xlsx_from_options(options)

    if out_csv_file:
        df.to_csv(out_csv_file, index= False)
        print('Exported to:', out_csv_file)

    if out_xls_file:
        if 'days' in df.columns:
            df['days'] = (df['days']/365).round(2)
        df = df.rename(columns= {
            'name': '基金经理',
            'company': '基金公司',
            'managers': '基金经理人数',
            'funds': '基金总数',
            'fund': '基金',
            'days': '管理时长',
            'years': '管理年限',
            'size': '基金规模',
            'best_return': '最佳回报',
            'annual': '年化收益',
        })
        df.to_excel(excel_writer= out_xls_file)
        print( tb.tabulate(df, headers='keys') )
        print('Exported to:', out_xls_file)

    return df

def cli_fund_read_fund_symbols(csv_file):
    df = pd.read_csv(csv_file, dtype=str)
    return df['symbol'].tolist() if ('symbol' in df) else []

# hiquant fund update <symbols>
# hiquant fund update <symbols.csv>
# hiquant fund update all
def cli_fund_update(params, options):
    if len(params) == 0:
        cli_fund_help()
        return

    df_fund_list = get_cn_fund_list(check_date= datetime_today())
    fund_symbol_names = dict_from_df(df_fund_list, 'symbol', 'name')

    if params[0] == 'all':
        params = df_fund_list['symbol'].tolist()
    elif params[0].endswith('.csv'):
        params = cli_fund_read_fund_symbols(params[0])
    elif ',' in params[0]:
        params = params[0].split(',')

    i = 0
    n = len(params)
    for param in params:
        i += 1
        name = param + ' - ' + fund_symbol_names[ param ]
        print('{}/{} - updating {} ...'.format(i, n, name))
        try:
            df = get_cn_fund_daily(symbol= param, check_date= datetime_today())
        except (KeyError, ValueError, IndexError) as err:
            print('error downloading', name)
            pass

    print('Done.')

def eval_fund_list(df_fund_list, date_from, date_to, alpha_base = None):
    fund_manager = get_fund_manager_mapping()
    days = (date_to - date_from).days
    eval_table = []
    for index, row in df_fund_list.iterrows():
        param = row['symbol']
        name = row['name']
        buy_state = row['buy_state']
        sell_state = row['sell_state']
        fee = row['fee']
        print('\r', index, '-', param, '-', name, '...', end='', flush= True)

        try:
            df = get_cn_fund_daily(symbol= param)
        except (KeyError, ValueError, IndexError) as err:
            print('\nerror downloading', name, ', skip.')
            continue

        fund_start = min(df.index)
        fund_days = (datetime_today() - fund_start).days
        #if fund_start > date_from:
        #    continue

        df = df[ df.index >= date_from ]
        df = df[ df.index < date_to ]

        try:
            df['pct_cum'] = (df['pct_change'] * 0.01 +1).cumprod()
            pct_cum = df['pct_cum'].iloc[-1] - 1.0
            pct_cum = round(pct_cum * 100, 2)
        except (KeyError, ValueError, IndexError) as err:
            #print('error calculating', param, name, ', skip.')
            continue

        if alpha_base is not None:
            df_base = get_index_daily( alpha_base )
            df_base = df_base[ df_base.index >= date_from ]
            df_base = df_base[ df_base.index < date_to ]
            df_base['pct_cum'] = df_base['close'] / df_base['close'].iloc[0]
            df['pct_cum'] -= df_base['pct_cum']

        risk_free_rate = 3.0 / 365
        daily_sharpe_ratio = (df['pct_change'].mean() - risk_free_rate) / df['pct_change'].std()
        sharpe_ratio = round(daily_sharpe_ratio * (252 ** 0.5), 2)

        max_drawdown = (1 - df['pct_cum'] / df['pct_cum'].cummax()).max()
        max_drawdown = round(100 * max_drawdown, 2)

        logreturns = np.diff( np.log(df['pct_cum']) )
        volatility = np.std(logreturns)
        annualVolatility = volatility * (252 ** 0.5)
        annualVolatility = round(annualVolatility * 100, 2)
        manager = ','.join(fund_manager[name]) if (name in fund_manager) else ''
        eval_table.append([param, name, manager, min(days, fund_days), pct_cum, sharpe_ratio, max_drawdown, annualVolatility, buy_state, sell_state, fee, fund_start, round(fund_days/365.0,1)])

    en_cols = ['symbol', 'name', 'manager', 'calc_days', 'pct_cum', 'sharpe', 'max_drawdown', 'volatility', 'buy_state', 'sell_state', 'fee', 'fund_start', 'fund_years']
    df = pd.DataFrame(eval_table, columns=en_cols)
    df['fund_start'] = df['fund_start'].dt.strftime('%Y-%m-%d')
    return df

# hiquant fund eval 002943 005669
# hiquant fund eval 002943 005669 -days=365
# hiquant fund eval data/myfunds.csv -days=365
# hiquant fund eval all -days=365 -sortby=sharpe -desc
def cli_fund_eval(params, options):
    if len(params) == 0:
        cli_fund_help()
        return

    df_fund_list = get_cn_fund_list()

    if params[0] == 'all':
        pass
    else:
        if params[0].endswith('.csv'):
            params = cli_fund_read_fund_symbols(params[0])
        elif ',' in params[0]:
            params = params[0].split(',')
        else:
            pass
        df_fund_list = df_fund_list[ df_fund_list['symbol'].isin(params) ]

    date_from, date_to, limit = date_limit_from_options(options)

    alpha_base = None
    yeartop = 0
    for k in options:
        if k.startswith('-alpha='):
            alpha_base = k.replace('-alpha=', '')
        if k.startswith('-yeartop='):
            yeartop = int(k.replace('-yeartop=', ''))
    
    df_fund_list = df_fund_list[ df_fund_list['buy_state'].isin(['限大额','开放申购']) ]

    if '-nc' in options:
        df_fund_list = df_fund_list[ ~ df_fund_list['name'].str.contains('C') ]
    if '-nlof' in options:
        df_fund_list = df_fund_list[ ~ df_fund_list['name'].str.contains('LOF') ]
    if '-netf' in options:
        df_fund_list = df_fund_list[ ~ df_fund_list['name'].str.contains('ETF') ]

    if yeartop > 0:
        symbols = []
        days = (date_to - date_from).days
        for i in range(0, days, 365):
            print('\ryear', i // 365)
            eval_from = date_from + dt.timedelta(days= i)
            df_eval = eval_fund_list(df_fund_list, date_from= eval_from, date_to= date_to)
            df_eval = df_eval.sort_values(by='pct_cum', ascending=False)
            if yeartop > 0:
                df_eval = df_eval.head(yeartop)
            symbols += df_eval['symbol'].tolist()
        df_fund_list = df_fund_list[ df_fund_list['symbol'].isin(set(symbols)) ]

    df_eval = eval_fund_list(df_fund_list, date_from= date_from, date_to= date_to, alpha_base= alpha_base)
    df_eval['annual'] = round((np.power((df_eval['pct_cum'] * 0.01 + 1), 1.0/(df_eval['calc_days']/365.0)) - 1.0) * 100.0, 1)
    df_eval['annual'] = df_eval[['pct_cum', 'annual']].min(axis= 1)

    if '-smart' in options:
        df_eval = filter_with_options(df_eval, options)
        df_eval['score'] = round(df_eval['pct_cum'] * df_eval['sharpe'] / df_eval['max_drawdown'], 2)
        df_eval = sort_with_options(df_eval, ['-sortby=score', '-desc'], by_default='score')
        if limit > 0:
            df_eval = df_eval.head(limit)
    else:
        df_eval = filter_with_options(df_eval, options)
        df_eval = sort_with_options(df_eval, options, by_default='pct_cum')
        if limit > 0:
            df_eval = df_eval.head(limit)

    print('\r', end= '', flush= True)
    print( tb.tabulate(df_eval, headers='keys') )

    out_csv_file, out_xls_file = csv_xlsx_from_options(options)

    if out_csv_file:
        df = df_eval[['symbol', 'name']]
        df.to_csv(out_csv_file, index= False)
        print('Exported to:', out_csv_file)
        print(df)

    if out_xls_file:
        years = round((date_to - date_from).days / 365.0, 1)
        df_eval = df_eval.drop(columns=['calc_days','sell_state','fund_start','volatility'])
        df_eval = df_eval.rename(columns= {
            'symbol': '基金代码',
            'name': '基金简称',
            'manager': '基金经理',
            'pct_cum': str(years) + '年收益率',
            'sharpe': '夏普比率',
            'max_drawdown': '最大回撤',
            'volatility': '波动率',
            'buy_state': '申购状态',
            'sell_state': '赎回状态',
            'fee': '手续费',
            'fund_start': '起始日期',
            'fund_years': '基金年数',
            'annual': '年化收益',
            'score': '综合评分',
        })
        df_eval.to_excel(excel_writer= out_xls_file)
        print( tb.tabulate(df_eval, headers='keys') )
        print('Exported to:', out_xls_file)

# hiquant fund plot 002943
# hiquant fund plot 002943 005669
# hiquant fund plot 002943 005669 -days=365
def cli_fund_plot(params, options, title= None):
    if len(params) == 0:
        cli_fund_help()
        return

    if params[0].endswith('.csv'):
        params = cli_fund_read_fund_symbols(params[0])
    elif ',' in params[0]:
        params = params[0].split(',')

    df_fund_list = get_cn_fund_list()
    fund_symbol_names = dict_from_df(df_fund_list, 'symbol', 'name')

    date_from, date_to, limit = date_limit_from_options(options)
    if limit == 0:
        limit = 100

    df_funds = None
    i = 0
    for param in params:
        if param in fund_symbol_names:
            name = param + ' - ' + fund_symbol_names[ param ]
        else:
            name = param
        print( name )
        
        i += 1
        if i > limit:
            break

        try:
            df = get_cn_fund_daily(symbol= param)
        except (KeyError, ValueError) as e:
            continue

        df = df[ df.index >= date_from ]
        df = df[ df.index < date_to ]
        df['pct_cum'] = round(((df['pct_change'] * 0.01 +1).cumprod() - 1.0) * 100.0, 1)
        if df_funds is None:
            df_funds = df[['pct_cum']]
            df_funds.columns = [ name ]
        else:
            df_funds[ name ] = df['pct_cum']
        #print(df)

    base = 'sh000300'
    for k in options:
        if k.startswith('-base='):
            base = k.replace('-base=', '')
    df_base = get_index_daily( base )
    df_base = df_base[ df_base.index >= date_from ]
    df_base = df_base[ df_base.index < date_to ]
    df_base['pct_cum'] = (df_base['close'] / df_base['close'].iloc[0] - 1.0) * 100.0

    if '-alpha' in options:
        table = []
        for k in df_funds.columns:
            df_funds[ k ] -= df_base['pct_cum']
            pct_change = df_funds[k].pct_change(1)
            daily_sharpe_ratio = pct_change.mean() / pct_change.std()
            sharpe_ratio = round(daily_sharpe_ratio * (252 ** 0.5), 2)
            table.append([k, sharpe_ratio])
        df = pd.DataFrame(table, columns=['fund', 'sharpe'])
        df = df.sort_values(by= 'sharpe', ascending= False)
        df_funds = df_funds[ df['fund'].tolist() ]
    else:
        index_symbol_names = dict_from_df( get_cn_index_list_df() )
        base_name = index_symbol_names[ base ] if base in index_symbol_names else base
        df_funds[ base_name ] = df_base['pct_cum']

    df_funds.index = df_funds.index.strftime('%Y-%m-%d')
    if '-mix' in options:
        df = df_funds[ [ base_name ] ]
        df_funds = df_funds.drop(columns=[ base_name ])
        df['mix'] = df_funds.mean(axis=1)
        df.plot(kind='line', ylabel='return (%)', figsize=(10,6), title= title)
    else:
        df_funds.plot(kind='line', ylabel='return (%)', figsize=(10,6), title= title)
    plt.xticks(rotation=15)
    plt.show()

    pass

def cli_fund_plot_man(params, options):
    df = cli_fund_manager(params, ['-s'])
    n = df.shape[0]
    if n == 0:
        print('Not found')
    elif n > 1:
        print('Found more than one, please specify exactly company and name, like 南方基金张磊')
    else:
        row = df.iloc[0]
        symbols = row['symbol']
        title = row['company'] + ' - ' + row['name']
        cli_fund_plot([symbols], options, title= title)

def cli_fund_backtest(params, options):
    if len(params) == 0:
        cli_fund_help()
        return

    date_from, date_to, limit = date_limit_from_options(options)
    if limit == 0:
        limit = 20

    base = 'sh000300'
    period = 365
    eval_period = 365 * 3
    for k in options:
        if k.startswith('-base='):
            base = k.replace('-base=', '')
        if k.startswith('-period='):
            period = int(k.replace('-period=', ''))
        if k.startswith('-eval='):
            eval_period = int(k.replace('-eval=', ''))

    df_fund_list = get_cn_fund_list()
    if params[0] == 'all':
        pass
    elif params[0].endswith('.csv'):
        params = cli_fund_read_fund_symbols(params[0])
        df_fund_list = df_fund_list[ df_fund_list['symbol'].isin(params) ]
    else:
        df_fund_list = df_fund_list[ df_fund_list['symbol'].isin(params) ]

    returns = []
    d = date_from
    while d < date_to:
        print('-'*20, d, '-'*20)
        df_eval = eval_fund_list(df_fund_list, (d - dt.timedelta(days= eval_period)), d)
        df_eval = sort_with_options(df_eval, ['-sortby=pct_cum','-desc'], by_default='sharpe').head(limit)
        print('\r', end= '', flush= True)
        print( tb.tabulate(df_eval, headers='keys') )

        d2 = min(d + dt.timedelta(days= period), date_to)
        n = df_eval.shape[0]
        period_pct_change = 0
        for i, row in df_eval.iterrows():
            symbol = row['symbol']
            df = get_cn_fund_daily(symbol)
            df = df[ df.index >= d ]
            df = df[ df.index < d2 ]
            if df.shape[0] < 1:
                continue
            df['pct_cum'] = ((df['pct_change'] * 0.01 +1).cumprod() - 1.0) * 100.0
            period_pct_change += df['pct_cum'].iloc[-1] / n

        period_pct_change -= 0.15 + 0.5 + 1.75 / 365 * period
        print(d, d2, 'return =', period_pct_change)
        returns.append([d, d2, period_pct_change])

        d += dt.timedelta(days= period)
        pass

    df_returns = pd.DataFrame(returns, columns=['date1', 'date2', 'pct_change'])
    df_returns.index = df_returns['date2']
    df_returns['pct_cum'] = ((df_returns['pct_change'] * 0.01 +1).cumprod() - 1.0) * 100.0

    df_base = get_index_daily( base )
    df_base = df_base[ df_base.index >= date_from ]
    df_base = df_base[ df_base.index < date_to ]
    df_base['pct_cum'] = (df_base['close'] / df_base['close'].iloc[0] - 1.0) * 100.0
    index_symbol_names = dict_from_df( get_cn_index_list_df() )
    base_name = index_symbol_names[ base ] if base in index_symbol_names else base
    df_returns[ base_name ] = df_base['pct_cum']

    print(df_returns)
    df_returns[['pct_cum', base_name]].plot(kind='line', ylabel='return (%)', figsize=(10,6))
    plt.show()

    pass

def cli_fund(params, options):
    if (len(params) == 0) or (params[0] == 'help'):
        cli_fund_help()
        return

    action = params[0]
    params = params[1:]

    if action == 'list':
        cli_fund_list(params, options)

    elif action == 'manager':
        cli_fund_manager(params, options)

    elif action == 'company':
        cli_fund_company(params, options)

    elif action == 'update':
        cli_fund_update(params, options)

    elif action in ['eval']:
        cli_fund_eval(params, options)

    elif action in ['backtest']:
        cli_fund_backtest(params, options)

    elif action in ['plot']:
        cli_fund_plot(params, options)

    elif action in ['show']:
        cli_fund_plot_man(params, options)

    else:
        print('invalid action:', action)
        cli_fund_help()



# -*- coding: utf-8 -*-

from .ma import SMA
from ..core.indicator_signal import register_signal_indicator

# --------------------------------------------
def _bias(close, length):
    return (close / SMA(close, length) - 1) * 100.0

# --------------------------------------------
def BIAS(close, fast = 6, middle = 12, slow = 24):
    return _bias(close,fast), _bias(close,middle), _bias(close,slow)

def signal_bias(df, inplace=False, fast=6, middle=12, slow=24):
    bfast, bmiddle, bslow = BIAS(df.close, fast=fast, middle=middle, slow=slow)
    if inplace:
        df['bias_fast'], df['bias_middle'], df['bias_slow'] = bfast, bmiddle, bslow
    buy_signal = (bmiddle < -8) | (bslow < -16)
    sell_signal = (bmiddle > 8) | (bslow > 16)
    return buy_signal.astype(int) - sell_signal.astype(int)

register_signal_indicator('bias', signal_bias, ['bias_fast','bias_middle','bias_slow'], 'BIAS', 'obos')



#!/usr/bin/env python
# -*- coding: utf-8; py-indent-offset:4 -*-
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import sys
import os
sys.path.append(os.getcwd())
sys.path.append(os.getcwd()+'/strategy')

import configparser

from ..version import __version__
from ..utils import parse_params_options, dict_from_config_items
from ..core import set_lang, get_hiquant_conf

from .cli_init import cli_init
from .cli_finance import cli_finance
from .cli_pepb import cli_pepb
from .cli_stockpool import cli_stockpool
from .cli_strategy import cli_strategy
from .cli_trade import cli_trade
from .cli_fund import cli_fund
from .cli_stock import cli_stock
from .cli_index import cli_index
from .cli_indicator import cli_indicator
from .cli_fundflow import cli_fundflow

def cli_main_help():
    syntax_tips = '''Syntax:
    __argv0__ <command>
    __argv0__ <module> <action> [options]

Global Commands:
    help ................... Get help for a command
    init ................. init a hiquant project

Modules:
    finance ................ Finance analysis on stock pool
    pepb ................... PE/PE analysis on stock pool
    stockpool .............. Create or edit stock pool .csv file
    strategy ............... Create or bench a strategy

    stock .................. Plot a stock with indicators
    index .................. Plot index with indicators
    fund ................... List or plot open funds
    indicator .............. List indicators
    fundflow ............... View main fundflow of stocks

Options:
    -v, --version .......... print out the version
    -d, --verbose .......... debug mode produces verbose log output

Example:
    __argv0__ init myFund
    cd myFund

    __argv0__ stock list us
    __argv0__ index list world

    __argv0__ stock plot AAPL -macd
    __argv0__ stock plot AAPL -ma -macd -kdj

    __argv0__ fund list
    __argv0__ fund plot 005669 000209

    __argv0__ stockpool create stockpool/mystocks.csv 600036 000002 600276 300357

    __argv0__ strategy create strategy/mystrategy.py
    ... modify my_strategy.py ...
    __argv0__ backtest strategy/mystrategy.py

    __argv0__ trade create etc/myfund.conf
    ... modify my_fund.conf ...
    __argv0__ backtrade etc/myfund.conf
    __argv0__ run etc/myfund.conf
'''.replace('__argv0__',os.path.basename(sys.argv[0]))

    print(syntax_tips)

def cli_main_params_options(params, options):
    # parse command line arguments
    cli_tools = {
        'init': cli_init,
        'finance': cli_finance,
        'pepb': cli_pepb,
        'stockpool': cli_stockpool,
        'strategy': cli_strategy,
        'trade': cli_trade,
        'fund': cli_fund,
        'stock': cli_stock,
        'index': cli_index,
        'indicator': cli_indicator,
        'fundflow': cli_fundflow,
    }

    if ('-v' in options) or ('--version' in options):
        print( __version__ )
        return

    if len(params) == 0:
        cli_main_help()
        return

    config = get_hiquant_conf()
    main_conf = dict_from_config_items(config.items('main'))
    if 'lang' in main_conf:
        set_lang(main_conf['lang'])
    else:
        lang = os.getenv('LANG')
        if lang is not None:
            set_lang(lang[:2])

    command = params[0]

    if command == 'help':
        if (len(params) > 1) and (params[1] in cli_tools):
                func = cli_tools[ params[1] ]
                func(['help'], options)
        elif (len(params) > 1) and (params[1] in ['backtest']):
                cli_strategy(['help'], options)
        elif (len(params) > 1) and (params[1] in ['backtrade', 'run']):
                cli_trade(['help'], options)
        else:
            cli_main_help()

    elif command == 'backtest':
        cli_strategy(params, options)

    elif command in ['backtrade', 'run']:
        cli_trade(params, options)

    elif command in cli_tools:
        func = cli_tools[ command ]
        func(params[1:], options)
    else:
        cli_main_help()

def cli_main():
    params, options = parse_params_options(sys.argv)
    cli_main_params_options(params, options)

if __name__ == "__main__":
    cli_main()



# -*- coding: utf-8 -*-

import pandas as pd
import numpy as np

from .basic import CROSS
from .ma import EMA
from ..core.indicator_signal import register_signal_indicator

# Parabolic Stop And Reverse (psar)
# --------------------------------------------
def SAR(close, high, low, af=0.02, max_af=0.2):
    af = float(af) if af and af > 0 else 0.02
    max_af = float(max_af) if max_af and max_af > 0 else 0.2

    # initialize
    m = high.shape[0]
    af0 = af
    bullish = True
    high_point = high.iloc[0]
    low_point = low.iloc[0]

    if close is not None:
        sar = close.copy()
    else:
        sar = low.copy()
        
    long = pd.Series(np.NAN, index=sar.index)
    short = long.copy()
    reversal = pd.Series(False, index=sar.index)
    _af = long.copy()
    _af.iloc[0:2] = af0

    # Calculate Result
    for i in range(2, m):
        reverse = False
        _af.iloc[i] = af

        if bullish:
            sar.iloc[i] = sar.iloc[i - 1] + af * (high_point - sar.iloc[i - 1])

            if low.iloc[i] < sar.iloc[i]:
                bullish, reverse, af = False, True, af0
                sar.iloc[i] = high_point
                low_point = low.iloc[i]
        else:
            sar.iloc[i] = sar.iloc[i - 1] + af * (low_point - sar.iloc[i - 1])

            if high.iloc[i] > sar.iloc[i]:
                bullish, reverse, af = True, True, af0
                sar.iloc[i] = low_point
                high_point = high.iloc[i]

        reversal.iloc[i] = reverse

        if not reverse:
            if bullish:
                if high.iloc[i] > high_point:
                    high_point = high.iloc[i]
                    af = min(af + af0, max_af)
                if low.iloc[i - 1] < sar.iloc[i]:
                    sar.iloc[i] = low.iloc[i - 1]
                if low.iloc[i - 2] < sar.iloc[i]:
                    sar.iloc[i] = low.iloc[i - 2]
            else:
                if low.iloc[i] < low_point:
                    low_point = low.iloc[i]
                    af = min(af + af0, max_af)
                if high.iloc[i - 1] > sar.iloc[i]:
                    sar.iloc[i] = high.iloc[i - 1]
                if high.iloc[i - 2] > sar.iloc[i]:
                    sar.iloc[i] = high.iloc[i - 2]

        if bullish:
            long.iloc[i] = sar.iloc[i]
        else:
            short.iloc[i] = sar.iloc[i]

    return long, short, _af, reversal

def signal_sar(df, inplace=False, af=0.02, max_af=0.2):
    long, short, af, reversal = SAR(df.close, df.high, df.low, af=af, max_af=max_af)
    if inplace:
        df['sar_long'], df['sar_short'] = long, short
    tmp = long.fillna(0) - short.fillna(0)
    tmp[ tmp > 0 ] = 1
    tmp[ tmp < 0 ] = 0
    return tmp.diff(1)

register_signal_indicator('sar', signal_sar, ['sar_long', 'sar_short'], 'SAR', 'trend')



# -*- coding: utf-8; py-indent-offset:4 -*-

import datetime as dt
import time
import random

from ..utils import datetime_today
from .fund import Fund

class Callback:
    context = None
    func = None
    called = False
    def __init__(self, func, context):
        self.context = context
        self.func = func
        self.called = False

class Trader:
    market = None
    date_start = None
    date_end = None

    timer_callbacks = {}
    bar_callbacks = []
    funds = []

    verbose = False

    def __init__(self, market):
        self.market = market
        self.date_start = market.date_start
        self.date_end = market.date_end
        self.funds = []
        self.bar_callbacks = []
        self.timer_callbacks = {}

    def set_verbose(self, verbose = True):
        self.verbose = verbose
        for fund in self.funds:
            fund.set_verbose(verbose)

    def run_daily(self, func, context, time='09:30'):
        time_only = dt.datetime.strptime(time, '%H:%M')
        time = time_only.strftime('%H:%M')

        if time in self.timer_callbacks:
            callback_list = self.timer_callbacks[ time ]
            callback_list.append( Callback(func, context) )
        else:
            self.timer_callbacks[ time ] = [ Callback(func, context) ]

    def run_on_bar_update(self, func, context):
        self.bar_callbacks.append( Callback(func, context) )

    def add_fund(self, fund):
        self.funds.append(fund)

    def run_fund(self, date_start, date_end, tick_period = 300):
        market = self.market
        self.date_start = date_start
        self.date_end = date_end

        market.current_time = market.current_date = date = date_start
        for fund in self.funds:
            fund.init_fund( date_end >= datetime_today() )

        while date < date_end:
            # print the date at bottom of screen, but do not change line
            #if not fund.verbose:
            print('\r... {} ...'.format(date.strftime('%Y-%m-%d %H:%M:%S')), end='', flush=True)

            next_date = min(date + dt.timedelta(days=1), date_end)

            for fund in self.funds:
                fund.before_day()

            # init timer for the day, convert str to real time value
            timers = {}
            for k in self.timer_callbacks:
                time_only = dt.datetime.strptime(k, '%H:%M')
                call_time = dt.datetime(date.year, date.month, date.day, time_only.hour, time_only.minute)
                callback_list = self.timer_callbacks[ k ]
                timers[ call_time ] = callback_list
                for callback in callback_list:
                    callback.called = False

            # if the day is in the past, it's a back-trade test
            if date < datetime_today():
                timer_keys = list(timers.keys())
                timer_keys.sort()
                for k in timer_keys:
                    for fund in self.funds:
                        fund.before_tick()

                    market.current_time = k
                    callback_list = timers[k]
                    for callback in callback_list:
                        if callback.called:
                            continue
                        func = callback.func
                        func( callback.context )
                        callback.called = True

                    for fund in self.funds:
                        fund.after_tick()
            else:
                # date == today, it's realtime trade
                # wait until this day really begin
                while dt.datetime.now() < date:
                    time.sleep(1)

                # sleep a random time within 10 min, to avoid two process downloading same file at same time
                n = random.randint(1,60) * 10
                time.sleep(n)

                # download the daily at idle time
                market.keep_daily_uptodate()

                market.current_time = dt.datetime.now()
                next_time_tick = market.current_time

                # if we run after market, we can update to get today's date
                market.update_daily_realtime( market.verbose )

                # loop until end of this day
                while market.current_time < next_date:
                    # if time to call tick
                    if market.current_time >= next_time_tick:
                        for fund in self.funds:
                            fund.before_tick()

                        if market.is_open():
                            # fetching the realtime price / volume data
                            data_updated = market.update_daily_realtime( market.verbose )
                            if data_updated:
                                # callback on bar update
                                for callback in self.bar_callbacks:
                                    func = callback.func
                                    func( callback.context )
                                pass

                        for fund in self.funds:
                            fund.after_tick()

                        next_time_tick += dt.timedelta(seconds = tick_period)

                    # callback triggered by timer
                    for k in timers:
                        if market.current_time >= k:
                            callback_list = timers[k]
                            for callback in callback_list:
                                if not callback.called:
                                    func = callback.func
                                    func( callback.context )
                                    callback.called = True

                    now_str = market.current_time.strftime('%Y-%m-%d %H:%M:%S')
                    print('\r... {} ...'.format(now_str), end = '', flush=True)
                    time.sleep(1)

                    market.current_time = dt.datetime.now()

            for fund in self.funds:
                fund.after_day()

            # move to next date
            market.current_date = date = next_date

        print('')
        pass

    def get_report(self):
        return [fund.get_report() for fund in self.funds]

    def print_report(self, report = None):
        if report is None:
            report = self.get_report()

        fund = self.funds[0] if (len(self.funds) > 0) else Fund(self.market, self)
        fund.print_report(report= report)

    def plot(self, report = None, compare_index = None, out_file= None):
        if report is None:
            report = self.get_report()

        fund = self.funds[0] if (len(self.funds) > 0) else Fund(self.market, self)
        fund.plot(report= report, compare_index= compare_index, out_file= out_file)



# -*- coding: utf-8; py-indent-offset:4 -*-

import os
import configparser

_hiquant_conf = {
}

def load_hiquant_conf():
    hiquant_conf_file = 'hiquant.conf'
    if os.path.isfile(hiquant_conf_file):
        config = configparser.ConfigParser()
        config.read(hiquant_conf_file, encoding='utf-8')
        _hiquant_conf['config'] = config
    else:
        print('file not found:', hiquant_conf_file)

def get_hiquant_conf():
    if 'config' not in _hiquant_conf:
        load_hiquant_conf()
    return _hiquant_conf['config']

def init_hiquant_conf(config_file):
    hiquant_conf_template = '''
[main]
lang = en
market = cn, hk, us

[order_cost]
close_tax = 0.001
open_commission = 0.0003
close_commission = 0.0003
min_commission = 5

[indicator]
# add your own indicators here
# my_abc = indicator/my_abc.py

[strategy]
# add your own strategy here
# 001 = strategy/001_macd.py
'''
    fp = open(config_file, 'w')
    fp.write(hiquant_conf_template)
    fp.close()



# -*- coding: utf-8; py-indent-offset:4 -*-

import os
import sys
import pandas as pd
import tabulate as tb

from ..utils import date_from_str
from ..core import get_all_signal_indicators
from ..core import Market, Stock

def cli_indicator_help():
    syntax_tips = '''Syntax:
    __argv0__ indicator list
    __argv0__ indicator bench <stockpool.csv> [<start>] [<end>] [<options>]

Actions:
    list ....................... list indicators supported in this tool
    bench ...................... bench indicators for the given stockpool

Symbols:
    <stockpool.csv> ................ stock pool csv file

Options:
    -rankby=final | overall ........ rank by final or overall performance
    -topN .......................... keep top N indicators, by default -top2

    -out=<out.csv> ................. export selected stocks into <out.csv> file

Example:
    __argv0__ indicator list
    __argv0__ indicator bench mystock.csv -top1 -out=mystock_idx.csv
'''.replace('__argv0__',os.path.basename(sys.argv[0]))

    print(syntax_tips)

def cli_indicator_list(params, options):
    indicators = get_all_signal_indicators()
    table = []
    for k, values in indicators.items():
        table.append([k, values['type'], values['label'], ', '.join(values['cols'])])
    df = pd.DataFrame(table, columns=['indicator', 'type', 'label', 'data'])
    print( tb.tabulate(df, headers='keys', tablefmt='psql') )

def cli_indicator_bench(params, options):
    if (len(params) == 0) or ('.csv' not in params[0]):
        print('\nError: A filename with .csv is expected.\n')
        return

    csv_file = params[0]
    stock_df = pd.read_csv(csv_file, dtype=str)

    start = params[1] if (len(params) > 1) else '3 years ago'
    end = params[2] if (len(params) > 2) else '1 week ago'
    date_start = date_from_str(start)
    date_end = date_from_str(end)

    rankby = 'overall' if '-rankby=overall' in options else 'final'

    topN = 2
    for k in options:
        if k.startswith('-top'):
            topN = int(k.replace('-top',''))

    symbols = stock_df['symbol'].tolist()
    if 'cash' in symbols:
        symbols.remove('cash')
    market = Market(date_start, date_end)
    market.watch(symbols)

    all_indicators = get_all_signal_indicators().keys()
    indicators = []
    n = stock_df.shape[0]
    for i, row in stock_df.iterrows():
        symbol = row['symbol']
        name = row['name']
        if symbol == 'cash':
            indicators.append('')
            continue
        df = market.get_daily(symbol, start=date_start, end=date_end)
        stock = Stock(symbol, name, df)
        stock.add_indicator(all_indicators, mix=False, inplace=False)
        rank_df = stock.rank_indicator(by = rankby)
        if '-v' in options:
            print('-' * 20, symbol, name, '-' * 20)
            print(df)
            print('-' * 60)
            print(rank_df)
            print('-' * 60)
        else:
            print('\r...', i+1, '/', n, '... ', end = '', flush = True)

        # find top indicators
        ranked_indicators = rank_df['indicator'].tolist()
        indicators.append( ' + '.join( ranked_indicators[:topN] ).replace('.','') )
    stock_df['indicators'] = indicators
    print('\n')

    print( tb.tabulate(stock_df, headers='keys', tablefmt='psql') )

    out_csv_file = ''
    for k in options:
        if k.startswith('-out=') and k.endswith('.csv'):
            out_csv_file = k.replace('-out=', '')
    if out_csv_file:
        stock_df.to_csv(out_csv_file, index= False)
        print('Exported to:', out_csv_file)

    print('')

def cli_indicator(params, options):
    if (len(params) == 0) or (params[0] == 'help'):
        cli_indicator_help()
        return

    action = params[0]
    params = params[1:]

    if action == 'list':
        cli_indicator_list(params, options)

    elif action == 'bench':
        cli_indicator_bench(params, options)

    else:
        print('invalid action:', action)





# -*- coding: utf-8; py-indent-offset:4 -*-

import os
import sys
import tabulate as tb

from ..utils import symbol_normalize, date_limit_from_options
from ..core import symbol_to_name, get_cn_stock_list_df, get_hk_stock_list_df, get_us_stock_list_df
from ..core import list_signal_indicators, get_order_cost
from ..core import Market, Stock

def cli_stock_help():
    syntax_tips = '''Syntax:
    __argv0__ stock list <market>
    __argv0__ stock plot <symbol> [<from_date>] [<to_date>] [<options>]

<market>
    cn ......................... China market
    hk ......................... Hong Kong market
    us ......................... USA market

<symbol> ....................... stock symbol, like 600036
<from_date> .................... default from 3 year ago
<to_date> ...................... default to yesterday

<options> ...................... options to plot the stock
    -ma, -macd, -kdj, etc. ..... add some indicators
    -all ....................... add all indicators
    -mix ....................... show trade result if mix trade signal of the indicators
    -top5 ...................... show only top 5 indicators according to result

Example:
    __argv0__ stock list cn
    __argv0__ stock plot 600036 -ma
    __argv0__ stock plot 600036 20180101 -ma -macd -kdj
    __argv0__ stock plot 600036 20180101 20210101 -ma -macd -kdj
    __argv0__ stock plot 600036 -ma -macd -mix
'''.replace('__argv0__',os.path.basename(sys.argv[0]))

    print( syntax_tips )

def cli_stock_list(params, options):
    list_funcs = {
        'cn': get_cn_stock_list_df,
        'us': get_us_stock_list_df,
        'hk': get_hk_stock_list_df,
    }
    if (len(params) == 0) or (not (params[0] in list_funcs)):
        print('Error: stock list expected param:', ', '.join(list(list_funcs.keys())))
        cli_stock_help()
        return
    func = list_funcs[ params[0] ]
    df = func()
    df['name'] = df['name'].str.slice(stop=60)
    print( tb.tabulate(df, headers='keys', showindex=False, tablefmt='psql') )
    print( 'Totally', df.shape[0], 'rows.\n')

def cli_stock_plot(params, options):
    symbol = symbol_normalize(params[0])
    name = symbol_to_name(symbol)

    date_start, date_end, limit = date_limit_from_options(options)

    market = Market(date_start, date_end)
    df = market.get_daily(symbol, start = date_start, end = date_end)
    stock = Stock(symbol, name, df)

    all_indicators = list_signal_indicators()

    indicators = []
    topN = 5
    inplace = True
    out_file = None
    for option in options:
        if option.startswith('-top'):
            topN = int(option.replace('-top',''))
        elif option.startswith('-out=') and (option.endswith('.png') or option.endswith('.jpg')):
            out_file = option.replace('-out=', '')
        else:
            k = option.replace('-','')
            if (k in all_indicators) and (not k in indicators):
                indicators.append(k)
    if '-all' in options:
        indicators = all_indicators
        inplace = False

    mix = '-mix' in options

    # add indicators and calculate performance
    stock.add_indicator(indicators, mix=mix, inplace= inplace, order_cost = get_order_cost())

    rank_df = stock.rank_indicator(by = 'final')
    if rank_df.shape[0] > 0:
        print(rank_df)

    # find top indicators
    ranked_indicators = rank_df['indicator'].tolist()
    other_indicators = ranked_indicators[topN:]
    df.drop(columns = other_indicators, inplace=True)

    stock.plot(out_file= out_file)

def cli_stock(params, options):
    if (len(params) == 0) or (params[0] == 'help'):
        cli_stock_help()
        return

    action = params[0]
    params = params[1:]

    if action == 'list':
        cli_stock_list(params, options)

    elif action in ['plot', 'view', 'show']:
        cli_stock_plot(params, options)

    else:
        print('invalid action:', action)
        cli_stock_help()



# -*- coding: utf-8; py-indent-offset:4 -*-

import time
import akshare as ak

def download_stock_pepb_history(symbol):
    print('\tfetching pe/pb data ...', symbol)
    df = ak.stock_a_lg_indicator(symbol)
    time.sleep(1)
    df = df.rename(columns={"trade_date":"date"}).sort_values(by='date')
    return df




# -*- coding: utf-8; py-indent-offset:4 -*-

# { df[key]: df[value] }
def dict_from_df(df, key_col = 'symbol', value_col = 'name'):
    mapping = {}
    for i, row in df.iterrows():
        mapping[ row[key_col] ] = row[value_col]
    return mapping

# -sortby=<col> -desc
def sort_with_options(df, options, by_default=''):
    sortby = by_default
    for k in options:
        if k.startswith('-sortby='):
            k = k.replace('-sortby=','')
            if k in df.columns:
                sortby = k
                break
    asc = '-desc' not in options
    return df.sort_values(by= sortby, ascending= asc).reset_index(drop= True) if sortby else df

# -<col>=<from>-<to>
def filter_with_options(df, options):
    for k in options:
        if k.startswith('-') and ('=' in k):
            k = k[1:]
            items = k.split('=')
            key = items[0]
            if (key in df.columns) and ('-' in items[1]):
                values = items[1].split('-')
                if values[0]:
                    df = df[ df[key] > float(values[0]) ]
                if values[1]:
                    df = df[ df[key] < float(values[1]) ]
    return df



#!/usr/bin/env python
# -*- coding: utf-8; py-indent-offset:4 -*-
###############################################################################
#
# Copyright (C) 2020-2021 Liming Xie
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
###############################################################################
import os.path
import codecs   # To use a consistent encoding
import setuptools

here = os.path.abspath(os.path.dirname(__file__))

# Get the long description from the relevant file
with codecs.open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

#from hiquant.version import __version__
lines = open(here + '/hiquant/version.py', mode='r').readlines()
for line in lines:
    if line.startswith('__version__ = '):
        __version__ = line.replace('__version__ = ','').split('\'')[1]
        break

# Package name
pname = 'hiquant'

# Generate links
gurl = 'https://github.com/floatinghotpot/' + pname
gdurl = gurl + '/tarball/' + __version__

setuptools.setup(
    name = pname,

    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version = __version__,
    description = 'Out-of-box python tool and framework for quantitative trading and study',
    long_description = long_description,
    long_description_content_type='text/markdown',

    # The project's main homepage.
    url = 'https://github.com/floatinghotpot/hiquant',
    download_url = 'https://github.com/floatinghotpot/hiquant',

    # Author details
    author = 'Liming Xie',
    author_email = 'liming.xie@gmail.com',

    # Choose your license
    license = 'GPLv3+',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 4 - Beta',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Intended Audience :: Financial and Insurance Industry',

        # Indicate which Topics are covered by the package
        'Topic :: Software Development',
        'Topic :: Office/Business :: Financial',

        # Pick your license as you wish (should match "license" above)
        ('License :: OSI Approved :: ' +
         'GNU General Public License v3 or later (GPLv3+)'),

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',

        # Operating Systems on which it runs
        'Operating System :: OS Independent',
    ],

    # What does your project relate to?
    keywords=[
        'stock',
        'fund',
        'finance',
        'spider',
        'quant',
        'quantitative',
        'investment',
        'trading',
        'algotrading',
        'data',
    ],

    platforms = 'any',
    python_requires='>=3.7',

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=setuptools.find_packages(exclude=['cache', 'output', 'log', 'samples', 'docs']),

    # List run-time dependencies here.
    # These will be installed by pip when your
    # project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires = [
      'requests',
      'pandas',
      'akshare',
      'tabulate',
      'matplotlib',
      'mplfinance',
    ],

    # List additional groups of dependencies here
    # (e.g. development dependencies).
    # You can install these using the following syntax, for example:
    # $ pip install -e .[dev,test]
    # extras_require={},

    # If there are data files included in your packages that need to be
    # installed, specify them here.  If using Python 2.6 or less, then these
    # have to be included in MANIFEST.in as well.
    # package_data={'sample': ['package_data.dat'],},

    # Although 'package_data' is the preferred approach, in some case you may
    # need to place data files outside of your packages. See:
    # http://docs.python.org/3.4/distutils/setupscript.html#installing-additional-files
    # In this case, 'data_file' will be installed into '<sys.prefix>/my_data'
    # data_files=[('my_data', ['data/data_file'])],

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    # entry_points={'console_scripts': ['sample=sample:main',],},
    entry_points={'console_scripts': ['hiquant=hiquant.cli:cli_main']},

    #scripts=['tools/hq.py'],
)



# -*- coding: utf-8; py-indent-offset:4 -*-

'''
symbol rule:

for China user (name in Chinese):
    China mainland market:
        stock: 6 digits, like 600036, 000002, 300357
        index: sh/sz + 6 digits, like sh000300
    Hongkong market:
        stock: 5 digits, like 00700, 09988
        any symbol like 'hk0700' or 'hk00700' will be converted to 00700

for US user (name in English):
    China mainland market:
        stock & index: 6 digits + exchange, like 600036.SS, 000002.SZ
    Hongkong market:
        stock: 4 digits + exchange, like 0700.HK
    US market:
        stock: like AAPL, AMZN, MSFT
        index: like ^GSPC

'''

def symbol_market( symbol ):
    if (len(symbol) == 6) and symbol[0].isdigit(): # 600036, 000002, 300357
        return 'cn'
    elif symbol.startswith('sh') or symbol.startswith('sz'): # sh000300, sz399001, index
        return 'cn'
    elif (len(symbol) == 5) and symbol[0].isdigit(): # 00700
        return 'hk'
    else:
        return 'us'

def symbol_normalize(symbol):
    if type(symbol) == str:
        if symbol[:2] in ['hk','HK','Hk','hK']:
            return '0' + symbol[-4:]
        else:
            return symbol
    elif type(symbol) == list:
        return [symbol_normalize(sym) for sym in symbol]
    else:
        raise ValueError('invalid symbol type: ' + type(symbol))

def symbol_yahoo_style(symbol):
    if (len(symbol) == 6) and symbol[0].isdigit():
        if symbol[0] in ['0','3']: # 000002, 300357
            return symbol + '.SZ'
        elif symbol[0] in ['6']: # 600036
            return symbol + '.SS'
        else:
            raise ValueError('invalid symbol: ' + symbol)
    elif symbol.startswith('sh'): # sh000300, shanghai index
        return symbol[-6:] + '.SS'
    elif symbol.startswith('sz'): # sz399001, shenzhen index
        return symbol[-6:] + '.SZ'
    elif (len(symbol) == 5) and symbol[0].isdigit(): # 00700
        return symbol[-4:] + '.HK'
    else:
        return symbol



# -*- coding: utf-8 -*-

from ..core.lang import add_lang

add_lang('en', {
    'buy': 'buy',
    'sell': 'sell',
    'stop loss': 'stop loss',
    'stop earn': 'stop earn',
    'signal': 'signal',
    'You can buy': 'You can buy',
    'You can sell': 'You can sell',
    'price': 'price',
    'volume': 'volume',
    'trade': 'trade',
    'position': 'position',
    'return(%)': 'return(%)',
    'position(%)': 'position(%)',
    'drawdown(%)': 'drawdown(%)',
    'quantitative backtesting': 'Quantitative backtesting',
})



# -*- coding: utf-8; py-indent-offset:4 -*-

import os
import sys

import pandas as pd
import tabulate as tb

from ..core import get_stockpool_df

def cli_stockpool_help():
    syntax_tips = '''Syntax:
    __argv0__ stockpool <action> <file.csv> [<symbols> | <another.csv>] [-out=<out.csv>]

Action:
    create .................... create a new stock pool csv file
    view ...................... show the content of the stock pool
    merge ..................... merge the stock pool with other symbols or another stock pool csv
    exclude ................... exclude symbols or another csv from this stock pool csv
    same ...................... find the same symbols existing in both csv file

Symbols:
    symbols ................... a list of stock symbols, like 600036 000002 300033
    file.csv .................. a csv file with stock symbol and names

Export:
    -out=<out.csv> ............ save result into a new csv file

Example:
    __argv0__ stockpool create mystocks.csv 600036 000002 300033
    __argv0__ stockpool view mystocks.csv
    __argv0__ stockpool exclude mystocks.csv 000651 600276 -out=mystocks.csv
    __argv0__ stockpool merge mystocks.csv file2.csv -out=mystocks.csv
'''.replace('__argv0__',os.path.basename(sys.argv[0]))

    print(syntax_tips)

def cli_stockpool_create(params, options):
    if (len(params) == 0) or ('.csv' not in params[0]):
        print('\nError: A filename with .csv is expected.\n')
        cli_stockpool_help()
        return

    csv_file = params[0]
    params = params[1:]

    if len(params) > 0:
        df = get_stockpool_df(params)
        df.to_csv(csv_file, index=False)
        print('Stockpool file created:', os.path.abspath(csv_file))
        print( tb.tabulate(df, headers='keys', tablefmt='psql') )
        print('')
    else:
        print('\nError: Symbol list is expected.\n')
        cli_stockpool_help()

def cli_stockpool(params, options):
    if (len(params) == 0) or (params[0] == 'help'):
        cli_stockpool_help()
        return

    action = params[0]
    params = params[1:]

    if (len(params) == 0) or ('.csv' not in params[0]):
        print('\nError: A filename with .csv is expected.\n')
        return

    csv_file = params[0]
    params = params[1:]

    # create csvfile with symbols
    if action == 'create':
        cli_stockpool_create(params, options)
        return

    # load the csv file
    df1 = pd.read_csv(csv_file, dtype=str)
    if action in ['view', 'show']:
        print( tb.tabulate(df1, headers='keys', tablefmt='psql') )
        return

    if (len(params) == 0):
        print('\nError: Symbols or another csv file is expected.\n')
        cli_stockpool_help()
        return

    # load or create the second csv file
    if '.csv' in params[0]:
        csv_file2 = params[0]
        df2 = pd.read_csv(csv_file2, dtype=str)
    else:
        df2 = get_stockpool_df( params )

    # now, merge, exclude, or find same
    if action == 'merge':
        df3 = pd.merge(df1, df2, on=['symbol','name'], how='outer').reset_index(drop=True)
    elif action == 'exclude':
        df3 = pd.concat([df1, df2, df2]).drop_duplicates(keep=False).reset_index(drop=True)
    elif action == 'same':
        df3 = pd.merge(df1, df2, on=['symbol','name'], how='inner').reset_index(drop=True)
    else:
        print('\nError: Unkown action:', action, '\n')
        cli_stockpool_help()
        return

    print( tb.tabulate(df3, headers='keys', tablefmt='psql') )

    for option in options:
        if ('-out=' in option) and option.endswith('.csv'):
            out_csv_file = option.replace('-out=','')
            df3.to_csv(out_csv_file, index=False)
            print('Saved to:', os.path.abspath(out_csv_file))

    print('')



import requests
import pandas as pd

from ..utils.symbol import symbol_normalize

def download_cn_stock_spot( symbols, verbose= False ):
    if len(symbols) > 100:
        df = pd.DataFrame()
        for i in range(0, len(symbols), 100):
            page = symbols[i:i+100]
            page_df = download_cn_stock_spot(page)
            df = df.append(page_df, ignore_index=True)
        return df
    else:
        sina_symbols = []
        for symbol in symbols:
            if (len(symbol) == 6) and symbol[0].isdigit():
                if symbol[0] == '6':
                    sina_symbols.append( 'sh' + symbol )
                elif symbol[0] in ['0','3']:
                    sina_symbols.append( 'sz' + symbol )
            elif (len(symbol) == 5) and symbol[0].isdigit():
                sina_symbols.append( 'hk' + symbol )
            elif symbol[:2].capitalize() == 'HK':
                sina_symbols.append( 'hk0' + symbol[-4:] )
            # ignore others, might invalid or index
            #else:
            #    sina_symbols.append( symbol )
        print('\rfetching quote data ...', end = '', flush = True)
        url = 'http://hq.sinajs.cn/list={0}'.format(','.join(sina_symbols))
        r = requests.get(url)
        print('\r', end = '', flush = True)

        table = []
        if(r.status_code == 200):
            lines = r.text.replace('var hq_str_','').replace('="',',').replace('";','').split('\n')
            for line in lines:
                v = line.split(',')
                symbol = v[0]
                symbol_prefix = symbol[:2]
                if symbol_prefix in ['sh', 'sz']:
                    # return data format:
                    # symbol = name, open, prevclose, lasttrade, high, low, buy, sell, volume, amount
                    # buy1-count, buy1-price, ...
                    # sell1-count, sell1-price, ...
                    # date, time
                    symbol = symbol[-6:]
                    # symbol, name, open, prevclose, lasttrade, high, low, volume, date, time
                    row = [symbol, v[1], v[2], v[3], v[4], v[5], v[6], v[9], v[31], v[32]]
                    table.append(row)
                elif symbol_prefix == 'hk':
                    # return data format:
                    # symbol = en_name, zh_name, open, prevclose, high, low, lasttrade,
                    # pricechange, changepercent, buy1-price, sell1-price, amount, volume, 0, 0, ?, ?,
                    # date, time
                    symbol = 'hk' + symbol[-4:]
                    # symbol, name, open, prevclose, lasttrade, high, low, volume, date, time
                    row = [symbol, v[2], v[3], v[4], v[5], v[6], v[7], v[13], v[18], v[19]]
                    table.append(row)
        df = pd.DataFrame(table, columns = ['symbol','name','open','prevclose','close','high','low','volume','date', 'time'])
        df = df.astype({
            'open':'float64',
            'prevclose':'float64',
            'close':'float64',
            'high':'float64',
            'low':'float64',
            'volume':'int64',
            'date':'datetime64'
        })
        df['symbol'] = symbol_normalize(df['symbol'].tolist())
        return df



# -*- coding: utf-8; py-indent-offset:4 -*-

import os
#import random
import datetime as dt
import pandas as pd

from ..utils import get_file_modify_time, datetime_today, symbol_normalize, symbol_market, dict_from_df

from ..data_source import *

def get_cached_download_df(csv_file, download_func, param = None, check_date = None):
    if type(param) == str:
        csv_file = csv_file.replace('{param}', param)

    # like ^GSPC.csv, will cause trouble when manually delete
    csv_file = csv_file.replace('^','_').replace('$','_')

    need_update = False
    if os.path.isfile(csv_file):
        if check_date is not None:
            modified_time = get_file_modify_time(csv_file)
            need_update = modified_time < check_date
        else:
            need_update = False
    else:
        need_update = True

    if need_update:
        for i in range(3):
            try:
                df = download_func(param)
            except (ValueError, IndexError) as err:
                _DOWNLOAD_RETRY_DELAY = 300
                print('downloading failed, try again after {} min'.format(_DOWNLOAD_RETRY_DELAY // 60))
                time.sleep(_DOWNLOAD_RETRY_DELAY)
                df = None

            if df is not None:
                df.to_csv(csv_file, index= bool(df.index.name))
                return df

        if df is None:
            print('downloading failed after 3 tries, loading cached data')

    df = pd.read_csv(csv_file, dtype=str)
    return df

def get_cn_stock_list_df():
    return get_cached_download_df('cache/cn_stock_list.csv', download_cn_stock_list, check_date= datetime_today())

def get_hk_stock_list_df():
    return get_cached_download_df('cache/hk_stock_list.csv', download_hk_stock_list, check_date= datetime_today())

def get_us_stock_list_df():
    return get_cached_download_df('cache/us_stock_list.csv', download_us_stock_list, check_date= datetime_today())

def get_cn_index_list_df():
    return get_cached_download_df('cache/cn_index_list.csv', download_cn_index_list, check_date= datetime_today())

def get_hk_index_list_df():
    return get_cached_download_df('cache/hk_index_list.csv', download_hk_index_list, check_date= datetime_today())

def get_us_index_list_df():
    return get_cached_download_df('cache/us_index_list.csv', download_us_index_list, check_date= datetime_today())

def get_world_index_list_df():
    return get_cached_download_df('cache/world_index_list.csv', download_world_index_list, check_date= datetime_today())

_market_funcs_get_stock_list_df = {
    'cn': get_cn_stock_list_df,
    'hk': get_hk_stock_list_df,
    'us': get_us_stock_list_df,
}

_market_funcs_get_index_list_df = {
    'cn': get_cn_index_list_df,
    'hk': get_hk_index_list_df,
    'us': get_us_index_list_df,
}

_market_funcs_download_index_daily_df = {
    'cn': download_cn_index_daily,
    'hk': download_cn_index_daily,
    'us': download_us_index_daily,
}

_enabled_markets = ['cn', 'hk', 'us']

def get_supported_market():
    return list(_market_funcs_get_stock_list_df.keys())

def enable_market(markets):
    if len(markets) == 0:
        raise ValueError('Must select at least one market')
    for market in markets:
        if market not in get_supported_market():
            raise ValueError('Market not supported yet: ' + market)
    _enabled_markets = markets

def get_enabled_market():
    return _enabled_markets

def get_all_stock_list_df():
    df = pd.DataFrame([],  columns=['symbol', 'name'])
    for market in _enabled_markets:
        func = _market_funcs_get_stock_list_df[ market ]
        market_df = func()[['symbol', 'name']]
        df = df.append(market_df, ignore_index=True)
    return df.reset_index(drop= True)

def get_all_index_list_df():
    df = pd.DataFrame([],  columns=['symbol', 'name'])
    df = df.append(get_world_index_list_df(), ignore_index= True)
    for market in _enabled_markets:
        func = _market_funcs_get_index_list_df[ market ]
        market_df = func()[['symbol', 'name']]
        df = df.append(market_df, ignore_index=True)
    return df.reset_index(drop= True)

def get_cn_stock_symbol_name():
    return dict_from_df(get_cn_stock_list_df(), 'symbol', 'name')

def get_all_symbol_name():
    symbol_name = dict_from_df(get_all_stock_list_df(), 'symbol', 'name')
    symbol_name.update(dict_from_df(get_all_index_list_df(), 'symbol', 'name'))
    return symbol_name

def get_symbol_name_dict(symbols):
    return dict_from_df(download_us_stock_quote(symbols), 'symbol', 'shortName')

def get_symbol_name(symbol):
    symbol_name = get_symbol_name_dict([symbol])
    return symbol_name[ symbol ] if (symbol in symbol_name) else symbol

def symbol_to_name(symbol):
    all_symbol_name = get_all_symbol_name()
    if type(symbol) == str:
        if symbol in all_symbol_name:
            return all_symbol_name[ symbol ]
        else:
            return get_symbol_name(symbol)
    elif type(symbol) == list:
        symbol_name = {}
        other_symbols = []
        for sym in symbol:
            if sym in all_symbol_name:
                symbol_name[ sym ] = all_symbol_name[ sym ]
            else:
                other_symbols.append( sym )
        if len(other_symbols) > 0:
            symbol_name.update( get_symbol_name_dict(other_symbols) )
        return symbol_name
    else:
        raise ValueError('invalid type for symbol: ' + type(symbol))

def get_stockpool_df(symbols):
    if type(symbols) == str:
        if symbols.endswith('.csv'):
            return pd.read_csv(symbols, dtype=str)
        else:
            symbols = symbols.replace(' ','').split(',')
    elif type(symbols) == list:
        pass
    else:
        raise ValueError('Exptected: symbols as .csv, str, or list')

    symbols = symbol_normalize(symbols)

    symbol_name = symbol_to_name(symbols)
    df = pd.DataFrame([], columns=['symbol','name'])
    df['symbol'] = list(symbol_name.keys())
    df['name'] = list(symbol_name.values())
    return df

def get_index_daily( symbol ):
    market = symbol_market( symbol )
    func = _market_funcs_download_index_daily_df[ market ]
    df = get_cached_download_df('cache/market/{param}_1d.csv', download_func= func, param= symbol, check_date= datetime_today())

    if 'date' in df.columns:
        df['date'] = pd.to_datetime(df['date'])
        df.set_index('date', inplace= True, drop= True)
        df = df.astype(float)

    return df

def get_daily( symbol ):
    # we get stock daily from yahoo, including us, cn, hk, etc.
    symbol = symbol_yahoo_style( symbol )
    func = download_us_stock_daily
    df = get_cached_download_df('cache/market/{param}_1d.csv', download_func= func, param= symbol, check_date= datetime_today())

    if 'date' in df.columns:
        df['date'] = pd.to_datetime(df['date'])
        df.set_index('date', inplace= True, drop= True)
        df = df.astype(float)

    return df

def get_market_fund_flow_daily():
    return get_cached_download_df('cache/market/cn_market_ff.csv', download_func= download_cn_market_fund_flow, param= None, check_date= datetime_today())

def get_stock_fund_flow_daily(symbol):
    market = symbol_market( symbol )
    if market == 'cn':
        df = get_cached_download_df('cache/market/{param}_ff.csv', download_func= download_cn_stock_fund_flow, param= symbol, check_date= datetime_today())
        if 'date' in df.columns:
            df['date'] = pd.to_datetime(df['date'])
            df.set_index('date', inplace= True, drop= True)
            df = df.astype(float)
        return df

    else:
        return None

def get_cn_stock_fund_flow_rank(days = '1d'):
    return download_cn_stock_fund_flow_rank(days)

def get_cn_sector_fund_flow_rank(days= '1d', sector= 'industry'):
    return download_cn_sector_fund_flow_rank(days, sector)

def get_stock_spot(symbols, verbose = False):
    cn_symbols = []
    us_symbols = []
    for symbol in symbols:
        market = symbol_market(symbol)
        if market in ['cn', 'hk']:
            cn_symbols.append( symbol )
        elif market in ['us']:
            us_symbols.append( symbol )
    df = pd.DataFrame([], columns=[
            'symbol', 'name',
            'open', 'prevclose',
            'close', 'high', 'low',
            'volume', 
            'date',
        ])
    if len(cn_symbols) > 0:
        cn_df = download_cn_stock_spot(cn_symbols, verbose)
        df = df.append(cn_df, ignore_index= True)
    if len(us_symbols) > 0:
        us_df = download_us_stock_spot(us_symbols, verbose)
        df = df.append(us_df, ignore_index= True)
    return df

def get_finance_balance_report(symbol, check_date= None):
    return get_cached_download_df('cache/finance/{param}_balance.csv', download_finance_balance_report, symbol, check_date= check_date)

def get_finance_income_report(symbol, check_date= None):
    return get_cached_download_df('cache/finance/{param}_income.csv', download_finance_income_report, symbol, check_date= check_date)

def get_finance_cashflow_report(symbol, check_date= None):
    return get_cached_download_df('cache/finance/{param}_cashflow.csv', download_finance_cashflow_report, symbol, check_date= check_date)

def get_finance_abstract_df(symbol, check_date= None):
    abstract_file = 'cache/finance/{}_abstract.csv'.format(symbol)

    if os.path.isfile(abstract_file) and (check_date is None):
        df = pd.read_csv(abstract_file, dtype= str)
        if 'date' in df.columns:
            df['date'] = pd.to_datetime(df['date'])
            df.set_index('date', inplace= True, drop= True)
            df = df.astype(float)
    else:
        balance_df = get_finance_balance_report(symbol, check_date= check_date)
        income_df = get_finance_income_report(symbol, check_date= check_date)
        cashflow_df = get_finance_cashflow_report(symbol, check_date= check_date)
        df = extract_abstract_from_report(symbol, balance_df, income_df, cashflow_df)
        df.to_csv(abstract_file, index= not ('date' in df.columns))

    return df

def get_ipoinfo_df(symbol, check_date= None):
    return get_cached_download_df('cache/finance/{param}_ipo.csv', download_ipo, symbol, check_date= check_date)

def get_dividend_history(symbol, check_date= None):
    return get_cached_download_df('cache/finance/{param}_dividend.csv', download_dividend_history, symbol, check_date= check_date)

def create_finance_indicator_df(check_date = None):
    print('Processing finance indicators for all stocks ...')
    symbol_name = get_cn_stock_symbol_name()
    symbols = list( symbol_name.keys() )
    symbols.sort()

    table = []
    cols = None
    i = 0
    n = len(symbols)
    for symbol in symbols:
        i += 1
        name = symbol_name[ symbol ]
        print('\r... {}/{} - {} {} ...'.format(i, n, symbol, name), end='', flush= True)

        abstract_df = get_finance_abstract_df(symbol, check_date= check_date)
        dividend_df = get_dividend_history(symbol, check_date= check_date)
        ipoinfo_df = get_ipoinfo_df(symbol, check_date= None)
        row = extract_finance_indicator_data(symbol, abstract_df, ipoinfo_df, dividend_df)

        if cols is None:
            cols = list( row.keys() )
            cols.append('name')

        row = list(row.values())
        row.append(name)

        table.append( row )

    print('')
    return pd.DataFrame(table, columns= cols)

def get_finance_indicator_df(symbols = None, check_date = None):
    df = get_cached_download_df('cache/cn_stock_indicator.csv', create_finance_indicator_df, param= check_date, check_date = check_date)
    df = df.astype({
        'ipo_years': 'float64',
        '3yr_grow_rate': 'float64',
        'grow_rate': 'float64',
        'grow': 'float64',
        'share_grow': 'float64',
        'start_bvps': 'float64',
        'now_bvps': 'float64',
        'eps': 'float64',
        'roe': 'float64',
        '3yr_roe': 'float64',
        'avg_roe': 'float64',
        'earn_speed': 'float64',
        'earn_yoy': 'float64',
        'debt_ratio': 'float64',
        'cash_ratio': 'float64',
        'earn_ttm': 'float64',
        'assets': 'float64',
        'equity': 'float64',
        'shares': 'float64',
    })
    if symbols is not None:
        df = df[ df['symbol'].isin(symbols) ].reset_index(drop= True)
    return df

def update_finance_indicator_df(symbols):
    indicator_df_file = 'cache/cn_stock_indicator.csv'
    if symbols is None:
        return
    else:
        for symbol in symbols:
            df = get_finance_abstract_df(symbol= symbol, check_date= datetime_today())
        if os.path.isfile(indicator_df_file):
            os.unlink(indicator_df_file)
        return get_finance_indicator_df(symbols)

def get_stock_pepb_history(symbol, check_date = None):
    df = get_cached_download_df('cache/pepb/{param}_pepb.csv', download_stock_pepb_history, param= symbol, check_date = check_date)
    if 'date' in df.columns:
        df['date'] = pd.to_datetime(df['date'])
        df.set_index('date', inplace= True, drop= True)
        df = df.astype(float)
    return df

def get_pepb_summary(symbol, check_date= None, years = 10):
    today = datetime_today()
    years_ago = today - dt.timedelta(days= 365 * years)
    df = get_stock_pepb_history(symbol, check_date = check_date)
    if df.shape[0] > 0:
        df = df.loc[ pd.to_datetime(years_ago) : pd.to_datetime(today) ]
        df['pe_ttm'].fillna(method='ffill', inplace=True)
        df['pe_ttm'].fillna(method='bfill', inplace=True)
        df['pb'].fillna(method='ffill', inplace=True)
        df['pb'].fillna(method='bfill', inplace=True)
        pe_list = df['pe_ttm'].tolist()
        pb_list = df['pb'].tolist()
    else:
        pe_list = [ 22.0 ]
        pb_list = [ 1.0 ]

    pe = pe_list[-1]
    pb = pb_list[-1]
    pe_pos = round(sum([1 if i < pe else 0 for i in pe_list]) * 100.0 / len(pe_list), 1)
    pb_pos = round(sum([1 if i < pb else 0 for i in pb_list]) * 100.0 / len(pb_list), 1)
    mean_pe = sum(pe_list) / len(pe_list)
    mean_pb = sum(pb_list) / len(pb_list)
    pe_ratio = round(pe / mean_pe, 3)
    pb_ratio = round(pe / mean_pb, 3)

    return {
        'symbol': symbol,
        'pe': pe,
        'pb': pb,
        'pe_pos': pe_pos,
        'pb_pos': pb_pos,
        'pe_ratio': pe_ratio,
        'pb_ratio': pb_ratio,
    }

def create_pepb_symmary_df(symbols= None, years = 10):
    print('Processing PE/PB summary for stocks ...')
    symbol_name = get_cn_stock_symbol_name()
    if symbols is None:
        symbols = list(symbol_name.keys())

    table = []
    cols = None
    for symbol in symbols:
        name = symbol_name[ symbol ]
        row = get_pepb_summary(symbol, check_date = datetime_today(), years= years)
        if cols is None:
            cols = list(row.keys())
            cols.append('name')
        row = list(row.values())
        row.append(name)
        table.append( row )

    return pd.DataFrame(table, columns=cols)

def get_pepb_symmary_df(symbols = None):
    if symbols is None:
        df = get_cached_download_df('cache/cn_stock_pepb.csv', create_pepb_symmary_df, check_date = datetime_today())
        df = df.astype({
            'pe': 'float64',
            'pb': 'float64',
            'pe_pos': 'float64',
            'pb_pos': 'float64',
            'pe_ratio': 'float64',
            'pb_ratio': 'float64',
        })
    else:
        df = create_pepb_symmary_df(symbols)

    return df

def get_macro_bank_interest_rate(country):
    df = get_cached_download_df('cache/bank/{paran}_macro_interest_rate.csv', download_macro_bank_interest_rate, param= country, check_date = datetime_today())
    df['date'] = pd.to_datetime(df.index.date)
    df.set_index('date', inplace=True, drop=True)
    return df

def get_cn_fund_list(check_date= None):
    df = get_cached_download_df('cache/cn_fund_list.csv', download_cn_fund_list, check_date= check_date)
    df.columns = ['symbol', 'name', 'val_1', 'cumval_1', 'val_2', 'cumval_2', 'val_change', 'pct_change', 'buy_state', 'sell_state', 'fee']
    return df

def get_cn_etf_list(check_date= None):
    df = get_cached_download_df('cache/cn_etf_list.csv', download_cn_etf_fund_list, check_date= check_date)
    df.columns = ['symbol', 'name', 'val_1', 'cumval_1', 'val_2', 'cumval_2', 'val_change', 'pct_change', 'buy_state', 'sell_state', 'fee']
    return df

def get_cn_fund_daily(symbol, check_date= None):
    df = get_cached_download_df('cache/fund/{param}_1d.csv', download_func= download_cn_fund_info, param= symbol, check_date= check_date)
    df.columns = ['date', 'value', 'pct_change']
    df = df.astype({
        'date': 'datetime64',
        'value': 'float64',
        'pct_change': 'float64',
    })
    df = df.set_index('date', drop= True)
    return df

def get_cn_etf_daily(symbol, check_date= None):
    df = get_cached_download_df('cache/fund/{param}_1d.csv', download_func= download_cn_etf_fund_info, param= symbol, check_date= check_date)
    df.columns = ['date', 'value', 'pct_change']
    df = df.astype({
        'date': 'datetime64',
        'value': 'float64',
        'pct_change': 'float64',
    })
    df = df.set_index('date', drop= True)
    return df

def get_cn_fund_manager(check_date= None):
    df = get_cached_download_df('cache/cn_fund_manager.csv', download_cn_fund_manager, check_date= check_date)
    df.columns = ['index', 'name', 'company', 'fund', 'days', 'size', 'best_return']
    df = df.drop(columns=['index'])
    df = df.astype({
        'days': 'float64',
        'size': 'float64',
        'best_return': 'float64',
    })
    return df



# -*- coding: utf-8; py-indent-offset:4 -*-

_lang_res = {
    '_current_lang': 'en'
    # 'en' = {},
    # 'zh' = {},
}

def add_lang(lang, translation):
    _lang_res[ lang ] = translation

def set_lang(lang = 'en'):
    if lang in _lang_res:
        _lang_res['_current_lang'] = lang

def get_lang():
    return _lang_res['_current_lang']

def LANG(key, lang = None):
    if lang is None:
        lang = _lang_res['_current_lang']

    if lang in _lang_res:
        lang_dict = _lang_res[ lang ]
        if key in lang_dict:
            return lang_dict[ key ]
        else:
            return key
    else:
        raise ValueError('Unknown language: ' + lang)



# -*- coding: utf-8 -*-

from .basic import REF, CROSS
from .ma import SMA, EMA
from ..core.indicator_signal import register_signal_indicator

# Triple ExponentiallySmoothed Average
# --------------------------------------------
def TRIX(close, length = 12, signal = 20):
    tr = EMA(EMA(EMA(close,length),length),length)
    ref = REF(tr,1)
    trix = (tr - ref) / ref * 100.0
    trma = SMA(trix, signal)
    return trix, trma
 
def signal_trix(df, inplace=False, length=10, signal=20):
    trix, trix_ma = TRIX(df.close, length=length, signal=signal)
    if inplace:
        df['trix'], df['trix_ma'] = trix, trix_ma
    return CROSS(trix, trix_ma)

register_signal_indicator('trix', signal_trix, ['trix', 'trix_ma'], 'TRIX', 'trend')
